@@ -14,7 +14,7 @@
 #include "pam.h"
 #include "mempool.h"
 
-LIQ_PRIVATE bool pam_computeacolorhash(struct acolorhash_table *acht, const rgba_pixel *const pixels[], unsigned int cols, unsigned int rows, const unsigned char *importance_map)
+LIQ_PRIVATE bool pam_computeacolorhash(struct acolorhash_table *acht, const liq_color *const pixels[], unsigned int cols, unsigned int rows, const unsigned char *importance_map)
 {
     const unsigned int ignorebits = acht->ignorebits;
     const unsigned int channel_mask = 255U>>ignorebits<<ignorebits;
@@ -175,14 +175,18 @@ LIQ_PRIVATE struct acolorhash_table *pam_allocacolorhash(unsigned int maxcolors,
     return t;
 }
 
-ALWAYS_INLINE static float pam_add_to_hist(const float *gamma_lut, hist_item *achv, unsigned int *j, const struct acolorhist_arr_item *entry, const float max_perceptual_weight)
+ALWAYS_INLINE static float pam_add_to_hist(struct temp_hist_item achv[], unsigned int *j, const struct acolorhist_arr_item *entry, const float max_perceptual_weight, int counts[])
 {
     if (entry->perceptual_weight == 0 && *j > 0) {
         return 0;
     }
-    const float w = MIN(entry->perceptual_weight/128.f, max_perceptual_weight);
-    achv[*j].adjusted_weight = achv[*j].perceptual_weight = w;
-    achv[*j].acolor = rgba_to_f(gamma_lut, entry->color.rgba);
+    const liq_color px = entry->color.rgba;
+    achv[*j].color = px;
+    const short cluster = ((px.r>>7)<<3) | ((px.g>>7)<<2) | ((px.b>>7)<<1) | (px.a>>7);
+    counts[cluster]++;
+    achv[*j].cluster = cluster;
+    const float w = MIN(entry->perceptual_weight/170.f, max_perceptual_weight);
+    achv[*j].weight = w;
     *j += 1;
     return w;
 }
@@ -199,8 +203,9 @@ LIQ_PRIVATE histogram *pam_acolorhashtoacolorhist(const struct acolorhash_table
     };
     if (!hist->achv) return NULL;
 
-    float gamma_lut[256];
-    to_f_set_gamma(gamma_lut, gamma);
+    /// Clusters form initial boxes for quantization, to ensure extreme colors are better represented
+    int counts[LIQ_MAXCLUSTER] = {};
+    struct temp_hist_item *temp = malloc(MAX(1, acht->colors) * sizeof(temp[0]));
 
     /* Limit perceptual weight to 1/10th of the image surface area to prevent
        a single color from dominating all others. */
@@ -211,26 +216,47 @@ LIQ_PRIVATE histogram *pam_acolorhashtoacolorhist(const struct acolorhash_table
     for(unsigned int i=0; i < acht->hash_size; ++i) {
         const struct acolorhist_arr_head *const achl = &acht->buckets[i];
         if (achl->used) {
-            total_weight += pam_add_to_hist(gamma_lut, hist->achv, &j, &achl->inline1, max_perceptual_weight);
+            total_weight += pam_add_to_hist(temp, &j, &achl->inline1, max_perceptual_weight, counts);
 
             if (achl->used > 1) {
-                total_weight += pam_add_to_hist(gamma_lut, hist->achv, &j, &achl->inline2, max_perceptual_weight);
+                total_weight += pam_add_to_hist(temp, &j, &achl->inline2, max_perceptual_weight, counts);
 
                 for(unsigned int k=0; k < achl->used-2; k++) {
-                    total_weight += pam_add_to_hist(gamma_lut, hist->achv, &j, &achl->other_items[k], max_perceptual_weight);
+                    total_weight += pam_add_to_hist(temp, &j, &achl->other_items[k], max_perceptual_weight, counts);
                 }
             }
         }
     }
+    hist->total_perceptual_weight = total_weight;
+
+    int begin = 0;
+    for(int i=0; i < LIQ_MAXCLUSTER; i++) {
+        hist->boxes[i].begin = begin;
+        hist->boxes[i].end = begin;
+        begin = begin + counts[i];
+    }
+
     hist->size = j;
     hist->total_perceptual_weight = total_weight;
     for(unsigned int k=0; k < hist->size; k++) {
         hist->achv[k].tmp.likely_colormap_index = 0;
     }
     if (!j) {
+        free(temp);
         pam_freeacolorhist(hist);
         return NULL;
     }
+
+    float gamma_lut[256];
+    to_f_set_gamma(gamma_lut, gamma);
+    for(int i=0; i < hist->size; i++) {
+        int j = hist->boxes[temp[i].cluster].end++;
+        hist->achv[j].acolor = rgba_to_f(gamma_lut, temp[i].color);
+        hist->achv[j].perceptual_weight = temp[i].weight;
+        hist->achv[j].adjusted_weight = temp[i].weight;
+    }
+    free(temp);
+
     return hist;
 }
 
@@ -248,7 +274,7 @@ LIQ_PRIVATE void pam_freeacolorhist(histogram *hist)
     hist->free(hist);
 }
 
-LIQ_PRIVATE colormap *pam_colormap(unsigned int colors, void* (*malloc)(size_t), void (*free)(void*))
+LIQ_PRIVATE LIQ_NONNULL colormap *pam_colormap(unsigned int colors, void* (*malloc)(size_t), void (*free)(void*))
 {
     assert(colors > 0 && colors < 65536);
 
@@ -286,3 +312,40 @@ LIQ_PRIVATE void to_f_set_gamma(float gamma_lut[], const double gamma)
     }
 }
 
+
+/* fixed colors are always included in the palette, so it would be wasteful to duplicate them in palette from histogram */
+LIQ_PRIVATE LIQ_NONNULL void remove_fixed_colors_from_histogram(histogram *hist, const int fixed_colors_count, const f_pixel fixed_colors[], const float target_mse)
+{
+    const float max_difference = MAX(target_mse/2.f, 2.f/256.f/256.f);
+    if (fixed_colors_count) {
+        for(int j=0; j < hist->size; j++) {
+            for(unsigned int i=0; i < fixed_colors_count; i++) {
+                if (colordifference(hist->achv[j].acolor, fixed_colors[i]) < max_difference) {
+                    hist->achv[j] = hist->achv[--hist->size]; // remove color from histogram by overwriting with the last entry
+                    j--; break; // continue searching histogram
+                }
+            }
+        }
+    }
+}
+
+LIQ_PRIVATE LIQ_NONNULL colormap *histogram_to_palette(const histogram *hist, void* (*malloc)(size_t), void (*free)(void*)) {
+    if (!hist->size) {
+        return NULL;
+    }
+    colormap *acolormap = pam_colormap(hist->size, malloc, free);
+    for(unsigned int i=0; i < hist->size; i++) {
+        acolormap->palette[i].acolor = hist->achv[i].acolor;
+        acolormap->palette[i].popularity = hist->achv[i].perceptual_weight;
+    }
+    return acolormap;
+}
+
+LIQ_PRIVATE LIQ_NONNULL void hist_reset_colors(const histogram *hist, const unsigned int colors) {
+    // likely_colormap_index (used and set in kmeans_do_iteration) can't point to index outside colormap
+    if (colors < 256) for(unsigned int j=0; j < hist->size; j++) {
+        if (hist->achv[j].tmp.likely_colormap_index >= colors) {
+            hist->achv[j].tmp.likely_colormap_index = 0; // actual value doesn't matter, as the guess is out of date anyway
+        }
+    }
+}
